#!/usr/bin/python
#
# Port Controler using UPnP.
#
# Copyright (c) 2005 Satoshi Fukutomi <fuktommy@inter7.jp>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

from time import sleep
from threading import Thread
from Tkinter import *

from upnp import findrouter

ports_size = 10
upnp_timeout = 10       # Seconds
save_status = 60        # Seconds
upnp_cycle = 1*60*60    # Seconds

router = None
openports = []

class Global(Frame):
    def createWidgets(self):
        self.LABEL = Label(self, text="Global IP Address")
        self.LABEL.pack(side=LEFT)

        self.ENTRY = Entry(self, width=20)
        self.ENTRY.pack(side=LEFT)

        self.GET = Button(self, text="Get", command=self.get_global)
        self.GET.pack(side=LEFT)

        if not router:
            for w in (self.ENTRY, self.GET):
                w.config(state=DISABLED)

    def __init__(self, *args):
        Frame.__init__(self, *args)
        self.pack(expand=True, fill=X)
        self.createWidgets()

    def get_global(self):
        wanipaddr = router.getwan()
        if wanipaddr:
            self.ENTRY.delete("0", END)
            self.ENTRY.insert(END, wanipaddr)
            self.master.status("Get global IP address succeeded")
        else:
            self.master.status("Get global IP address failed")


class PortOpener(Thread):
    def __init__(self, port, protocol, status):
        Thread.__init__(self)
        self.port = port
        self.protocol = protocol
        self.status = status

    def run(self):
        while True:
            sleep(upnp_cycle)
            if (self.port, self.protocol) not in openports:
                break
            flag = router.openport(self.port,
                                   self.protocol,
                                   "%s%d" % (self.protocol, self.port))
            if flag:
                self.status("Open port %d/%s succeeded" %
                            (self.port, self.protocol.lower()))
            else:
                self.master.status("Open port %d/%s failed" %
                                   (self.port, self.protocol.lower()))


class CleanStatus(Thread):
    def __init__(self, status):
        Thread.__init__(self)
        self.status = status

    def run(self):
        sleep(save_status)
        self.status.config(state=NORMAL)
        self.status.delete("0", END)
        self.status.config(state=DISABLED)


class PortPanel(Frame):
    def createWidgets(self):
        self.LABEL = Label(self, text="Port")
        self.LABEL.pack(side=LEFT)

        self.ENTRY = Entry(self, width=20)
        self.ENTRY.pack(side=LEFT)

        self.PROTOCOL = StringVar()
        r = {}
        for p in ("TCP", "UDP"):
            r[p] = Radiobutton(self, text=p, variable=self.PROTOCOL, value=p)
            r[p].pack(side=LEFT)
        r["TCP"].select()

        self.OPEN = Button(self, text="Open", command=self.open_port)
        self.OPEN.pack(side=LEFT)

        self.CLOSE = Button(self, text="Close", command=self.close_port,
                            state=DISABLED)
        self.CLOSE.pack(side=LEFT)

        if not router:
            for w in (self.ENTRY, self.OPEN):
                w.config(state=DISABLED)
            for k in r:
                r[k].deselect()
                r[k].config(state=DISABLED)


    def __init__(self, *args):
        Frame.__init__(self, *args)
        self.pack(expand=True, fill=X)
        self.createWidgets()

    def open_port(self):
        try:
            port = int(self.ENTRY.get())
            protocol = self.PROTOCOL.get()
            flag = router.openport(port, protocol, "%s%d" % (protocol, port))
            if flag:
                openports.append((port, protocol))
                po = PortOpener(port, protocol, self.master.status)
                po.setDaemon(True)
                po.start()
                self.master.status("Open port %d/%s succeeded" % 
                                   (port, protocol.lower()))
                self.OPEN.config(state=DISABLED)
                self.ENTRY.config(state=DISABLED)
                self.CLOSE.config(state=NORMAL)
            else:
                self.master.status("Open port %d/%s failed" %
                                   (port, protocol.lower()))
        except ValueError:
            self.master.status("Bad Number")

    def close_port(self):
        try:
            port = int(self.ENTRY.get())
            protocol = self.PROTOCOL.get()
            flag = router.closeport(port, protocol)
            if flag:
                openports.remove((port, protocol))
                self.master.status("Close port %d/%s succeeded" % 
                                   (port, protocol.lower()))
                self.CLOSE.config(state=DISABLED)
                self.ENTRY.config(state=NORMAL)
                self.OPEN.config(state=NORMAL)
            else:
                self.master.status("Close port %d/%s failed" %
                                   (port, protocol.lower()))
        except ValueError:
            self.master.status("Bad Number")


class PortCtrl(Frame):
    def createWidgets(self):
        self.master.title(u"PortCtrl")

        self.TOP = Frame(self)
        self.TOP.pack(side=TOP, expand=True)
        self.ROUTER = Label(self.TOP, text="Router: %s" % router)
        self.ROUTER.pack(side=LEFT)

        self.GLOBAL = Global(self)
        self.GLOBAL.pack()

        self.PANEL = []
        for i in range(ports_size):
            self.PANEL.append(PortPanel(self))
            self.PANEL[i].pack()

        self.BOTTOM = Frame(self)
        self.BOTTOM.pack()
        self.QUIT = Button(self.BOTTOM, text="Quit", command=sys.exit)
        self.QUIT.pack(side=LEFT)
        self.QUIT_DESC = Label(self.BOTTOM, text="(Close all ports)")
        self.QUIT_DESC.pack(side=LEFT)

        self.STATUS = Entry(self, width=30)
        if router:
            self.status("Status: OK")
        else:
            self.status("Router not found.")
        self.STATUS.pack(side=BOTTOM, fill=X)

    def __init__(self, master=None):
        Frame.__init__(self, master)
        self.pack(expand=True, fill=BOTH)
        self.createWidgets()

    def status(self, message):
        self.STATUS.config(state=NORMAL)
        self.STATUS.delete("0", END)
        self.STATUS.insert(END, message)
        self.STATUS.config(state=DISABLED)
        cs = CleanStatus(self.STATUS)
        cs.setDaemon(True)
        cs.start()


def main():
    try:
        try:
            global router
            router = findrouter(timeout=upnp_timeout)
            app = PortCtrl()
            app.mainloop()
        finally:
            for port, protocol in openports:
                router.closeport(port, protocol)
    except KeyboardInterrupt:
        sys.exit()

if __name__ == "__main__":
    main()
